#ifndef JOB_SCHEDULING_H
#define JOB_SCHEDULING_H

#include <string>
#include <vector>

/// Data in The Standard Workload Format version 2.2
/// http://www.cs.huji.ac.il/labs/parallel/workload/swf.html
struct JobData
{
    /// 1. Job Number -- a counter field, starting from 1.
    int jobNumber;

    /// 2. Submit Time -- in seconds. The earliest time the log refers to is zero, and is the submittal time the of the first job.
    double submitTime;
    /// 3. Wait Time -- in seconds. The difference between the job's submit time and the time at which it actually began to run.
    double waitTime;
    /// 4. Run Time -- in seconds. The wall clock time the job was running (end time minus start time).
    double runTime;
    ///Note that when values are rounded to an integral number of seconds (as often happens in logs) a run time of 0 is possible
    /// and means the job ran for less than 0.5 seconds. On the other hand it is permissable to use floating point values for time fields.

    /// 5. Number of Allocated Processors.
    int  numberOfAllocatedProcessors;

    /// 6. Average CPU Time Used -- both user and system, in seconds. This is the average over all processors of the CPU time used, and
    /// may therefore be smaller than the wall clock runtime.
    double averageCPUTimeUsed;

    /// 7. Used Memory -- in kilobytes. This is again the average per processor.
    double usedMemory;

    /// 8. Requested Number of Processors.
    int requestedNumberOfProcessors;

    /// 9. Requested Time. This can be either runtime (measured in wallclock seconds), or average CPU time per processor (also in seconds)
    /// -- the exact meaning is determined by a header comment. In many logs this field is used for the user runtime estimate (or upper bound) used in backfilling. If a log contains a request for total CPU time, it is divided by the number of requested processors.
    double requestedTime;  // requires modification?

    /// 10. Requested Memory (again kilobytes per processor).
    double requestedMemory;

    /// 11. Status -- 1 if the job was completed, 0 if it failed, and 5 if cancelled. If information about chekcpointing or swapping is included,
    /// other values are also possible. See usage note below.
    int status;

    enum JobStatus{ FAILED=1, SUCCESSFUL=0 };

    /// 12. User ID -- a natural number, between one and the number of different users.
    int userID;

    /// 13. Group ID -- a natural number, between one and the number of different groups.
    int groupID;

    /// 14. Executable (Application) Number -- a natural number, between one and the number of different applications appearing in the workload.
    /// In some logs, this might represent a script file used to run jobs rather than the executable directly; this should be noted in a header comment.
    int applicationNumber;

    /// 15. Queue Number -- a natural number, between one and the number of different queues in the system. The nature of the system's queues
    /// should be explained in a header comment. This field is where batch and interactive jobs should be differentiated: we suggest the
    /// convention of denoting interactive jobs by 0.
    int queueNumber;

    /// 16. Partition Number -- a natural number, between one and the number of different partitions in the systems. The nature of the system's partitions
    ///  should be explained in a header comment. For example, it is possible to use partition numbers to identify which machine in a cluster was used.
    int partitionNumber;

    /// 17. Preceding Job Number -- this is the number of a previous job in the workload, such that the current job can only start after the termination
    /// of this preceding job. Together with the next field, this allows the workload to include feedback as described below.
    int precedingJobNumber;

    /// 18. Think Time from Preceding Job -- this is the number of seconds that should elapse between the termination of the preceding job and the submittal of this one.
    double thinkTimeFromPrecedingJob;
};

/*
Usage of the Status field
The main usage of the status field is to note the job's status. This isn't as straightforward as it sounds.

The simple case is jobs that complete normally, and have status 1.

The harder case is jobs that don't complete normally. This can happen for several reasons:

   1. The job failed (e.g. segmentation fault). This is given status 0.
   2. The job was canceled by the user (like ^C in Unix). This is given status 5. Note that cancelled jobs may have positive runtimes and processors if cancelled after they started to run, or 0 or -1 if cancelled while waiting in the queue.
   3. The job was killed by the system (e.g. because it exceeded its requested run time). This may be given different status values in different logs; it will typically be 0 or 5, but might also be 1.

Note also that the distinction between failure / cancellation / killing is not necessarily accurate, as the distinction typically does not appear in the original logs.

If a log contains information about checkpoints and swapping out of jobs, a job can have multiple lines in the log. In fact, we propose that the job information appear twice. First, there will be one line that summarizes the whole job: its submit time is the submit time of the job, its runtime is the sum of all partial runtimes, and its code is 0 or 1 according to the completion status of the whole job. In addition, there will be separate lines for each instance of partial execution between being swapped out. All these lines have the same job ID and appear consecutively in the log. Only the first has a submit time; the rest only have a wait time since the previous burst. The completed code for all these lines is 2, meaning ``to be continued''; the completion code for the last such line is 3 or 4, corresponding to completion or being killed. It should be noted that such details are only useful for studying the behavior of the logged system, and are not a feature of the workload. Such studies should ignore lines with completion codes of 0 and 1, and only use lines with 2, 3, and 4. For workload studies, only the single-line summary of the job should be used, as identified by a code of 0 or 1.

To summarize, the status field codes are (or should be) as follows:
0	Job Failed
1	Job completed successfully
2	This partial execution will be continued
3	This is the last partial execution, job completed
4	This is the last partial execution, job failed
5	Job was cancelled (either before starting or during run)
*/


bool parseJobLine(std::string line, JobData &jobData);


/// Returns a vector of jobs from file fileName.
/// Jobs are sorted by user id (and submit time).
std::vector<JobData> parseFile(std::string fileName);

/// Returns a vector of jobs without the partial jobs (jobs with statuses 2, 3 and 4).
/// Only 0 (failed), 1 (successful) and 5 (canceled) are left.
std::vector<JobData> filterOutPartialJobs(const std::vector<JobData> &jobs);

struct Exception : public std::exception
{
   std::string ex;
   Exception(std::string str) : ex(str) {}
   virtual ~Exception() throw() {};
   const char* what() const throw() { return ex.c_str(); }
};



/// Attributes
typedef std::vector<double> (*AttributeFunction)(const std::vector<JobData> &, int);
std::vector<double> submitTime(const std::vector<JobData> &allJobs, int jobIndex);
std::vector<double> waitTime(const std::vector<JobData> &allJobs, int jobIndex);
std::vector<double> status(const std::vector<JobData> &allJobs, int jobIndex);
std::vector<double> requestedNumberOfProcessors(const std::vector<JobData> &allJobs, int jobIndex);
std::vector<double> requestedTime(const std::vector<JobData> &allJobs, int jobIndex);
std::vector<double> requestedMemory(const std::vector<JobData> &allJobs, int jobIndex);
std::vector<double> userID(const std::vector<JobData> &allJobs, int jobIndex);
std::vector<double> lastStatusOfUser(const std::vector<JobData> &allJobs, int jobIndex);
std::vector<double> lastStatusOfApplication(const std::vector<JobData> &allJobs, int jobIndex);
std::vector<double> lastStatusOfUserWithApplication(const std::vector<JobData> &allJobs, int jobIndex);
std::vector<double> lastJobOfUserWithApplicationWasCanceled(const std::vector<JobData> &allJobs, int jobIndex);
std::vector<double> lastRunTimeOfUserWithApplication(const std::vector<JobData> &allJobs, int jobIndex);
std::vector<double> lastAverageCPUTimeOfUserWithApplication(const std::vector<JobData> &allJobs, int jobIndex);
std::vector<double> elapsedTimeSinceUserLastUsedApplication(const std::vector<JobData> &allJobs, int jobIndex);
std::vector<double> elapsedTimeSinceUserLastUsedCluster(const std::vector<JobData> &allJobs, int jobIndex);


double jobStatusNeuronValue(int status);


void createTrainingFile(const std::vector<JobData> &allJobs, const std::vector<AttributeFunction> &inputAttributes, const std::vector<AttributeFunction> &outputAttributes, std::string outputFileName, double trainingSetSize = 50, double validationSetSize = 25, double testSetSize = 25);
int changeTrainingSamplePositivesRatio(std::vector<JobData> &allJobs, int trainingAndValidationSize, double positivesRatio);
void createUserIdLookupTable(const std::vector<JobData> &allJobs);
std::vector<double> createTrainingSample(const std::vector<JobData> &allJobs, int jobIndex, const std::vector<AttributeFunction> &inputAttributes, const std::vector<AttributeFunction> &outputAttributes);
/// Returns the number of input values (neurons).
int inputSize(const std::vector<JobData> &allJobs, const std::vector<AttributeFunction> &inputAttributes);
int outputSize(const std::vector<JobData> &allJobs, const std::vector<AttributeFunction> &outputAttributes);
std::vector<double> createAttributeValues(int numberOfNeurons, int activeNeuron);



#endif // JOB_SCHEDULING_H
