#ifndef NN_H
#define NN_H


#include <vector>
#include <string>

#include <CL/opencl.h>

#include "PSO.h"

#define KERNEL_FILE_NAME "NN_kernels.cl"


/// perhaps split into NN.h/cpp , NN_PSO.h/cpp, NN_BP.h/cpp

float sigmoid(float x);

// abstract base class
// Reads in neural network configuration data (layers) and samples.
class NN
{
public:
    /// Trains the neural network with the method depending on the derived class.
    /// Returns elapsed training time in seconds (doesn't count preprocessing).
    virtual double train() = 0;
    std::vector<float> operator() (std::vector<float> input);

    /// test functions

    /// Returns the mean square error of the test set.
    float testMSE();

    /// Returns the true positive rate of the test set. TPR = TP / P = TP / (TP + FN)
    float testTPR(int *P = NULL, int *N = NULL, int *TP = NULL, int *FP = NULL, int *TN = NULL, int *FN = NULL);

    void saveWeightsToFile(std::string fileName);
    void makeLiftChart(std::string fileName);


    // Runs training for numIterations iterations and returns average time for one iteration in seconds.
    // Erases previous training.
    virtual double timing(int numIterations = 10) = 0;

    virtual void printInfo();

    NN(std::string nnetConfigFile, std::string trainingDataFile);
    NN(); // for NN_reloaded
    virtual ~NN();

    /// Returns number of input neurons.
    int getInputSize();
    /// Returns number of output neurons.
    int getOutputSize();
    /// Returns the size of the layer. 0 = input; numLayers-1 == output
    int getLayerSize(int layer);
    /// Returns total number of weights in a network.
    int getNumWeights();
    /// Returns number of synapses of the network
    int getNumSynapses();
    /// Returns number of layers in a neural network
    int getNumLayers();
    /// Returns number of neurons in a neural network (including input and output neurons).
    int getNumNeurons();
    /// Returns size of widest layer in a network.
    int getWidestLayer();
    /// Returns number of training samples (input-ouput pairs).
    int getNumTrainingSamples();
    /// Returns number of validation samples (input-ouput pairs).
    int getNumValidationSamples();
    /// Returns number of test samples (input-ouput pairs).
    int getNumTestSamples();
    /// Returns number of samples for a given input vector.
    int getNumInputSamples(const std::vector<float> &input);
    /// Returns number of samples for a given output vector.
    int getNumOutputSamples(const std::vector<float> &output);

    // Calculates NN's output for input.
    void simulateNeuralNetwork(float *input, float *weights, float *output);
    std::vector<float> simulateNeuralNetwork(std::vector<float> input);


    // for timing purposes only (leaves only size samples);
    void downsizeTrainingSamples(int size);

private:
    void readInNNetConfig(std::string fileName);
    void readInTrainingData(std::string fileName);

    void shuffleSamples(std::vector<float> &inputs, std::vector<float> &targets, int inputSize, int targetSize);



protected:
    /// For the given inputs and targets calculate mean squared error.
    float computeMSE(std::vector<float> &input, std::vector<float> &target);

    std::vector<float> scaleInputData(std::vector<float> in);
    //void inverseScaleOutputData(float const* original_out, float * scaled_out);
    std::vector<float> inverseScaleOutputData(const std::vector<float> &out);
    void propagateLayer(float *input, float *weights, float *output,int previous_layer_size,int layer_size);

public:
    // host vectors
    std::vector<float> h_weights_;
protected:
    std::vector<float> h_training_inputs_;
    std::vector<float> h_validation_inputs_;
    std::vector<float> h_test_inputs_;
    std::vector<float> h_training_targets_; //expected outputs
    std::vector<float> h_validation_targets_;
    std::vector<float> h_test_targets_;
    std::vector<float> h_outputs_;   //calculated outputs

    std::vector<float> in_ranges_; // inputs are scaled when read in (from file, and from operator() )
    std::vector<float> h_out_ranges_; // [[0],[1]] - range for first output; [[2],[3]] - 2nd output...
    std::vector<int> h_layers_; //number of neurons in each layer [i,h1,..,hn,o]


    // just for CPU nnet simulation
    float *h_output_;
    float *h_temp_output_;

};


/// Neural network trained with particle swarm optimization without OpenCL.
/// Reads in PSO parameters
class NN_PSO : virtual public NN
{
public:
    NN_PSO(std::string nnetConfigFile, std::string trainingDataFile, std::string psoConfigFile);
    virtual ~NN_PSO();

    virtual double train();

    virtual void printInfo();

    virtual double timing(int numIterations = 10);

    /// Returns number of different weight sets there are.
    int getNumberOfNetworks();
    /// Returns how many network simulations are preformed.
    int getNumSimulations();

protected:
    PsoParameters psoParameters_;

    int bestParticleIndex;

    /// calculates outputs for every input and every network and stores outputs in h_outputs_
    virtual void simulateNeuralNetworks(std::vector<float> &input);

    virtual std::vector<float> computeMSEs(const std::vector<float> &targets);



};

// Common OpenCL class.
// Creates context, device, command queue, program. Contains device type data.
// Copies samples, NN layers and output ranges to device memory.
class NN_OpenCL : virtual public NN
{
public:
    NN_OpenCL(std::string nnetConfigFile, std::string trainingDataFile, cl_device_type device_type);
    virtual ~NN_OpenCL();
    virtual void printInfo();

protected:

    cl_context context_;
    cl_device_id device_;
    cl_command_queue queue_;
    cl_program program_;
    cl_kernel sim_kernel_;

    cl_device_type device_type_;

    cl_mem d_training_inputs_;
    cl_mem d_validation_inputs_;
    cl_mem d_test_inputs_;
    cl_mem d_training_targets_; //expected outputs
    cl_mem d_validation_targets_;
    cl_mem d_test_targets_;

    cl_mem vectorToClmem(const std::vector<float> &vec);

    cl_mem d_outputs_;

    cl_mem d_out_ranges_;
    cl_mem d_layers_;

    cl_mem d_weights_;
};

// Common PSO OpenCL class.
// Contains weights, outputs and MSEs on device.
// Simulated NN with kernels created in derived classes.
class NN_PSO_OpenCL : public NN_PSO, public NN_OpenCL
{
public:
    NN_PSO_OpenCL(std::string nnetConfigFile, std::string trainingDataFile, std::string psoConfigFile, cl_device_type device_type);
    virtual ~NN_PSO_OpenCL();
    virtual void printInfo();

    virtual void simulateNeuralNetworks(std::vector<float> &input);

protected:
    // Returns a maximum work group size for simulation (which mostly depends on local memory usage).
    virtual int getMaxWorkGroupSize() = 0;

    virtual std::vector<float> computeMSEs(const std::vector<float> &targets);

    int mseGroupSize_;
    int simGroupSize_;



    // openCL variables
    cl_kernel mse_kernel_;

    cl_mem d_MSEs_;
};

// Faster version (for GPU). Neural network is simulated in local memory.
// Weights (for one network) and simulation buffers (2 for each thread, size of widest layer) must fit into local memory.
class NN_PSO_OpenCL_local : public NN_PSO_OpenCL
{
public:
    NN_PSO_OpenCL_local(std::string nnetConfigFile, std::string trainingDataFile, std::string psoConfigFile, cl_device_type device_type);
    virtual ~NN_PSO_OpenCL_local();
    virtual void printInfo();
private:
    // Returns a maximum work group size (which mostly depends on local memory usage).
    int getMaxWorkGroupSize();
};

// GPU version of the local memory version.
class NN_PSO_OpenCL_GPU_local : public NN_PSO_OpenCL_local
{
public:
    NN_PSO_OpenCL_GPU_local(std::string nnetConfigFile, std::string trainingDataFile, std::string psoConfigFile);
};

// CPU version of the local memory version.
class NN_PSO_OpenCL_CPU_local : public NN_PSO_OpenCL_local
{
public:
    NN_PSO_OpenCL_CPU_local(std::string nnetConfigFile, std::string trainingDataFile, std::string psoConfigFile);
};


// Slower version. Better for CPU. Neural network is simulated in private/global memory.
class NN_PSO_OpenCL_private : public NN_PSO_OpenCL
{
public:
    NN_PSO_OpenCL_private(std::string nnetConfigFile, std::string trainingDataFile, std::string psoConfigFile, cl_device_type device_type);
    virtual ~NN_PSO_OpenCL_private();
    virtual void printInfo();
private:
    int getMaxWorkGroupSize();
};

// GPU version of the local memory version.
class NN_PSO_OpenCL_GPU_private : public NN_PSO_OpenCL_private
{
public:
    NN_PSO_OpenCL_GPU_private(std::string nnetConfigFile, std::string trainingDataFile, std::string psoConfigFile);
};

// CPU version of the local memory version.
class NN_PSO_OpenCL_CPU_private : public NN_PSO_OpenCL_private
{
public:
    NN_PSO_OpenCL_CPU_private(std::string nnetConfigFile, std::string trainingDataFile, std::string psoConfigFile);
};



// Backpropagation


class BpParameters
{
public:
    BpParameters(std::string fileName);
    BpParameters();

    void PrintInfo();

    float learningRate;
    int stripLength; // calculate validation after every stripLength training calculations

    // maybe add inertia?

    // weight limit for initialization
    float lowerLimit, upperLimit;

    // terminating conditions
    int iterations;
    float maxFitness;
    float time;

private:

    void SetDefaults();

};


// NN_BP

class NN_BP : virtual public NN
{
public:
    NN_BP(std::string nnetConfigFile, std::string trainingDataFile, std::string bpConfigFile);
    virtual ~NN_BP();

    virtual double train();

    virtual double timing(int numIterations = 10);

    virtual void printInfo();

    /// Calculates the output for the sample from inputSamples with index sampleIndex and stores it in h_outputs_
    virtual void simulateNeuralNetwork(std::vector<float> &inputSamples, int sampleIndex);

    /// Compute the mean squared error of all the samples.
    virtual float computeMSE(std::vector<float> &inputSamples, std::vector<float> &sampleTargets);


protected:
    // BP parameters
    BpParameters bpParameters_;

    // Calcualtes output, but also outputs from all neurons in all layers.
    // Outputs from output neurons are not scaled in *all_outputs (ie. are in the range [0,1]), but in *output are scaled.
    void simulateNeuralNetworkForBP(float *input, float *weights, float *output, float *all_outputs);


    // used in training();
    virtual void initializeWeights();
    virtual void initializeDeltaWeights();
    // Calculates outputs of all neurons for the given input. The final outputs are not scale (are in the range [0,1]).
    virtual void propagateForward(std::vector<float> &inputSamples, int sampleIndex);
    virtual void propagateBackwards(std::vector<float> &outputSamples, int sampleIndex);
    virtual void applyDeltaWeights();
    virtual void finalizeTraining();


    // just for CPU BP simulation
    float *h_deltas_;
    float *h_prev_deltas_; //
    float *h_all_outputs_; // output from each neuron
    float *h_delta_weights_;

    /// Returns output of the neuron with the index index from the layer layer.
    float getOutput(float *all_outputs, int layer, int index);

    /// Returns a pointer to the first output(neuron) from the layer layer.
    float* getLayerOutput(float *all_outputs, int layer);

    /// Returns a pointer to the weights on the link to the input of the neuron. First weight is the bias.
    /// Input neurons don't have any weights, so layer [1, numLayers-1]. neuron e= [0,layerSize-1].
    float* getNeuronWeights(float *weights, int layer, int neuron);
};

class NN_BP_OpenCL : public NN_BP, public NN_OpenCL
{
public:
    NN_BP_OpenCL(std::string nnetConfigFile, std::string trainingDataFile, std::string bpConfigFile, cl_device_type device_type);
    virtual ~NN_BP_OpenCL();

    virtual void printInfo();

    virtual void simulateNeuralNetwork(std::vector<float> &inputSamples, int sampleIndex);

private:

    // just for device BP simulation
    cl_mem d_all_outputs_; // output from each neuron
    cl_mem d_prev_deltas_;
    cl_mem d_deltas_;
    cl_mem d_delta_weights_;
    cl_mem d_output_;
    cl_mem d_temp_output_;

    // kernels
    cl_kernel layer_propagate_forward_kernel_;
    cl_kernel layer_propagate_backwards_kernel_;
    cl_kernel apply_delta_weights_kernel_;


    // used in training();
    virtual void initializeWeights();
    virtual void initializeDeltaWeights();
    virtual void propagateForward(std::vector<float> &inputSamples, int sampleIndex);
    virtual void propagateBackwards(std::vector<float> &outputSamples, int sampleIndex);
    virtual void applyDeltaWeights();
};

// GPU version of the BP NN.
class NN_BP_OpenCL_GPU : public NN_BP_OpenCL
{
public:
    NN_BP_OpenCL_GPU(std::string nnetConfigFile, std::string trainingDataFile, std::string bpConfigFile);
    virtual ~NN_BP_OpenCL_GPU() {}
};

// CPU version of the BP NN.
class NN_BP_OpenCL_CPU : public NN_BP_OpenCL
{
public:
    NN_BP_OpenCL_CPU(std::string nnetConfigFile, std::string trainingDataFile, std::string bpConfigFile);
    virtual ~NN_BP_OpenCL_CPU() {}
};


// NN_reloaded

// Uses weights loaded from a file. Can't be trained.
class NN_reloaded : public NN
{
public:
    NN_reloaded(std::string weightsFile);
    virtual double train();
    virtual double timing(int numIterations = 10);
    virtual void printInfo();
};


#endif // NN_H
